@@ -1,3 +1,5 @@
+from __future__ import annotations
+
 import logging
 import re
 from typing import Any
@@ -10,6 +12,7 @@ from typing import TYPE_CHECKING
 from typing import Union
 
 from sqlalchemy import Column
+from sqlalchemy import literal_column
 from sqlalchemy import Numeric
 from sqlalchemy import text
 from sqlalchemy import types as sqltypes
@@ -38,7 +41,6 @@ from ..operations import ops
 from ..operations import schemaobj
 from ..operations.base import BatchOperations
 from ..operations.base import Operations
-from ..util import compat
 from ..util import sqla_compat
 
 if TYPE_CHECKING:
@@ -111,24 +113,26 @@ class PostgresqlImpl(DefaultImpl):
         if defaults_equal:
             return False
 
-        if None in (conn_col_default, rendered_metadata_default):
+        if None in (
+            conn_col_default,
+            rendered_metadata_default,
+            metadata_column.server_default,
+        ):
             return not defaults_equal
 
-        # check for unquoted string and quote for PG String types
-        if (
-            not isinstance(inspector_column.type, Numeric)
-            and metadata_column.server_default is not None
-            and isinstance(
-                metadata_column.server_default.arg, compat.string_types
-            )
-            and not re.match(r"^'.*'$", rendered_metadata_default)
-        ):
-            rendered_metadata_default = "'%s'" % rendered_metadata_default
+        metadata_default = metadata_column.server_default.arg
 
+        if isinstance(metadata_default, str):
+            if not isinstance(inspector_column.type, Numeric):
+                metadata_default = re.sub(r"^'|'$", "", metadata_default)
+                metadata_default = f"'{metadata_default}'"
+
+            metadata_default = literal_column(metadata_default)
+
+        # run a real compare against the server
         return not self.connection.scalar(
-            text(
-                "SELECT %s = %s"
-                % (conn_col_default, rendered_metadata_default)
+            sqla_compat._select(
+                literal_column(conn_col_default) == metadata_default
             )
         )
 
@@ -137,16 +141,16 @@ class PostgresqlImpl(DefaultImpl):
         table_name: str,
         column_name: str,
         nullable: Optional[bool] = None,
-        server_default: Union["_ServerDefault", "Literal[False]"] = False,
+        server_default: Union[_ServerDefault, Literal[False]] = False,
         name: Optional[str] = None,
-        type_: Optional["TypeEngine"] = None,
+        type_: Optional[TypeEngine] = None,
         schema: Optional[str] = None,
         autoincrement: Optional[bool] = None,
-        existing_type: Optional["TypeEngine"] = None,
-        existing_server_default: Optional["_ServerDefault"] = None,
+        existing_type: Optional[TypeEngine] = None,
+        existing_server_default: Optional[_ServerDefault] = None,
         existing_nullable: Optional[bool] = None,
         existing_autoincrement: Optional[bool] = None,
-        **kw: Any
+        **kw: Any,
     ) -> None:
 
         using = kw.pop("postgresql_using", None)
@@ -170,7 +174,7 @@ class PostgresqlImpl(DefaultImpl):
                 )
             )
 
-        super(PostgresqlImpl, self).alter_column(
+        super().alter_column(
             table_name,
             column_name,
             nullable=nullable,
@@ -182,7 +186,7 @@ class PostgresqlImpl(DefaultImpl):
             existing_server_default=existing_server_default,
             existing_nullable=existing_nullable,
             existing_autoincrement=existing_autoincrement,
-            **kw
+            **kw,
         )
 
     def autogen_column_reflect(self, inspector, table, column_info):
@@ -231,13 +235,13 @@ class PostgresqlImpl(DefaultImpl):
         metadata_indexes,
     ):
 
-        conn_indexes_by_name = dict((c.name, c) for c in conn_indexes)
+        conn_indexes_by_name = {c.name: c for c in conn_indexes}
 
-        doubled_constraints = set(
+        doubled_constraints = {
             index
             for index in conn_indexes
             if index.info.get("duplicates_constraint")
-        )
+        }
 
         for ix in doubled_constraints:
             conn_indexes.remove(ix)
@@ -250,15 +254,19 @@ class PostgresqlImpl(DefaultImpl):
                 while isinstance(expr, UnaryExpression):
                     expr = expr.element
                 if not isinstance(expr, Column):
+                    if sqla_compat.sqla_2:
+                        msg = ""
+                    else:
+                        msg = "; not supported by SQLAlchemy reflection"
                     util.warn(
-                        "autogenerate skipping functional index %s; "
-                        "not supported by SQLAlchemy reflection" % idx.name
+                        "autogenerate skipping functional index "
+                        f"{idx.name!r}{msg}"
                     )
                     metadata_indexes.discard(idx)
 
     def render_type(
-        self, type_: "TypeEngine", autogen_context: "AutogenContext"
-    ) -> Union[str, "Literal[False]"]:
+        self, type_: TypeEngine, autogen_context: AutogenContext
+    ) -> Union[str, Literal[False]]:
         mod = type(type_).__module__
         if not mod.startswith("sqlalchemy.dialects.postgresql"):
             return False
@@ -270,7 +278,7 @@ class PostgresqlImpl(DefaultImpl):
         return False
 
     def _render_HSTORE_type(
-        self, type_: "HSTORE", autogen_context: "AutogenContext"
+        self, type_: HSTORE, autogen_context: AutogenContext
     ) -> str:
         return cast(
             str,
@@ -280,7 +288,7 @@ class PostgresqlImpl(DefaultImpl):
         )
 
     def _render_ARRAY_type(
-        self, type_: "ARRAY", autogen_context: "AutogenContext"
+        self, type_: ARRAY, autogen_context: AutogenContext
     ) -> str:
         return cast(
             str,
@@ -290,7 +298,7 @@ class PostgresqlImpl(DefaultImpl):
         )
 
     def _render_JSON_type(
-        self, type_: "JSON", autogen_context: "AutogenContext"
+        self, type_: JSON, autogen_context: AutogenContext
     ) -> str:
         return cast(
             str,
@@ -300,7 +308,7 @@ class PostgresqlImpl(DefaultImpl):
         )
 
     def _render_JSONB_type(
-        self, type_: "JSONB", autogen_context: "AutogenContext"
+        self, type_: JSONB, autogen_context: AutogenContext
     ) -> str:
         return cast(
             str,
@@ -312,17 +320,17 @@ class PostgresqlImpl(DefaultImpl):
 
 class PostgresqlColumnType(AlterColumn):
     def __init__(
-        self, name: str, column_name: str, type_: "TypeEngine", **kw
+        self, name: str, column_name: str, type_: TypeEngine, **kw
     ) -> None:
         using = kw.pop("using", None)
-        super(PostgresqlColumnType, self).__init__(name, column_name, **kw)
+        super().__init__(name, column_name, **kw)
         self.type_ = sqltypes.to_instance(type_)
         self.using = using
 
 
 @compiles(RenameTable, "postgresql")
 def visit_rename_table(
-    element: RenameTable, compiler: "PGDDLCompiler", **kw
+    element: RenameTable, compiler: PGDDLCompiler, **kw
 ) -> str:
     return "%s RENAME TO %s" % (
         alter_table(compiler, element.table_name, element.schema),
@@ -332,7 +340,7 @@ def visit_rename_table(
 
 @compiles(PostgresqlColumnType, "postgresql")
 def visit_column_type(
-    element: PostgresqlColumnType, compiler: "PGDDLCompiler", **kw
+    element: PostgresqlColumnType, compiler: PGDDLCompiler, **kw
 ) -> str:
     return "%s %s %s %s" % (
         alter_table(compiler, element.table_name, element.schema),
@@ -344,7 +352,7 @@ def visit_column_type(
 
 @compiles(ColumnComment, "postgresql")
 def visit_column_comment(
-    element: "ColumnComment", compiler: "PGDDLCompiler", **kw
+    element: ColumnComment, compiler: PGDDLCompiler, **kw
 ) -> str:
     ddl = "COMMENT ON COLUMN {table_name}.{column_name} IS {comment}"
     comment = (
@@ -366,7 +374,7 @@ def visit_column_comment(
 
 @compiles(IdentityColumnDefault, "postgresql")
 def visit_identity_column(
-    element: "IdentityColumnDefault", compiler: "PGDDLCompiler", **kw
+    element: IdentityColumnDefault, compiler: PGDDLCompiler, **kw
 ):
     text = "%s %s " % (
         alter_table(compiler, element.table_name, element.schema),
@@ -412,15 +420,15 @@ class CreateExcludeConstraintOp(ops.AddConstraintOp):
     def __init__(
         self,
         constraint_name: Optional[str],
-        table_name: Union[str, "quoted_name"],
+        table_name: Union[str, quoted_name],
         elements: Union[
             Sequence[Tuple[str, str]],
-            Sequence[Tuple["ColumnClause", str]],
+            Sequence[Tuple[ColumnClause, str]],
         ],
-        where: Optional[Union["BinaryExpression", str]] = None,
+        where: Optional[Union[BinaryExpression, str]] = None,
         schema: Optional[str] = None,
-        _orig_constraint: Optional["ExcludeConstraint"] = None,
-        **kw
+        _orig_constraint: Optional[ExcludeConstraint] = None,
+        **kw,
     ) -> None:
         self.constraint_name = constraint_name
         self.table_name = table_name
@@ -432,8 +440,8 @@ class CreateExcludeConstraintOp(ops.AddConstraintOp):
 
     @classmethod
     def from_constraint(  # type:ignore[override]
-        cls, constraint: "ExcludeConstraint"
-    ) -> "CreateExcludeConstraintOp":
+        cls, constraint: ExcludeConstraint
+    ) -> CreateExcludeConstraintOp:
         constraint_table = sqla_compat._table_for_constraint(constraint)
 
         return cls(
@@ -452,8 +460,8 @@ class CreateExcludeConstraintOp(ops.AddConstraintOp):
         )
 
     def to_constraint(
-        self, migration_context: Optional["MigrationContext"] = None
-    ) -> "ExcludeConstraint":
+        self, migration_context: Optional[MigrationContext] = None
+    ) -> ExcludeConstraint:
         if self._orig_constraint is not None:
             return self._orig_constraint
         schema_obj = schemaobj.SchemaObjects(migration_context)
@@ -462,7 +470,7 @@ class CreateExcludeConstraintOp(ops.AddConstraintOp):
             *self.elements,
             name=self.constraint_name,
             where=self.where,
-            **self.kw
+            **self.kw,
         )
         for (
             expr,
@@ -476,12 +484,12 @@ class CreateExcludeConstraintOp(ops.AddConstraintOp):
     @classmethod
     def create_exclude_constraint(
         cls,
-        operations: "Operations",
+        operations: Operations,
         constraint_name: str,
         table_name: str,
         *elements: Any,
-        **kw: Any
-    ) -> Optional["Table"]:
+        **kw: Any,
+    ) -> Optional[Table]:
         """Issue an alter to create an EXCLUDE constraint using the
         current migration context.
 
@@ -543,16 +551,16 @@ class CreateExcludeConstraintOp(ops.AddConstraintOp):
 
 @render.renderers.dispatch_for(CreateExcludeConstraintOp)
 def _add_exclude_constraint(
-    autogen_context: "AutogenContext", op: "CreateExcludeConstraintOp"
+    autogen_context: AutogenContext, op: CreateExcludeConstraintOp
 ) -> str:
     return _exclude_constraint(op.to_constraint(), autogen_context, alter=True)
 
 
 @render._constraint_renderers.dispatch_for(ExcludeConstraint)
 def _render_inline_exclude_constraint(
-    constraint: "ExcludeConstraint",
-    autogen_context: "AutogenContext",
-    namespace_metadata: "MetaData",
+    constraint: ExcludeConstraint,
+    autogen_context: AutogenContext,
+    namespace_metadata: MetaData,
 ) -> str:
     rendered = render._user_defined_render(
         "exclude", constraint, autogen_context
@@ -563,7 +571,7 @@ def _render_inline_exclude_constraint(
     return _exclude_constraint(constraint, autogen_context, False)
 
 
-def _postgresql_autogenerate_prefix(autogen_context: "AutogenContext") -> str:
+def _postgresql_autogenerate_prefix(autogen_context: AutogenContext) -> str:
 
     imports = autogen_context.imports
     if imports is not None:
@@ -572,8 +580,8 @@ def _postgresql_autogenerate_prefix(autogen_context: "AutogenContext") -> str:
 
 
 def _exclude_constraint(
-    constraint: "ExcludeConstraint",
-    autogen_context: "AutogenContext",
+    constraint: ExcludeConstraint,
+    autogen_context: AutogenContext,
     alter: bool,
 ) -> str:
     opts: List[Tuple[str, Union[quoted_name, str, _f_name, None]]] = []
@@ -642,7 +650,7 @@ def _exclude_constraint(
 
 
 def _render_potential_column(
-    value: Union["ColumnClause", "Column"], autogen_context: "AutogenContext"
+    value: Union[ColumnClause, Column], autogen_context: AutogenContext
 ) -> str:
     if isinstance(value, ColumnClause):
         template = "%(prefix)scolumn(%(name)r)"
